@@ -1314,6 +1314,16 @@ void _glfwPlatformUpdateIMEState(_GLFWwindow *w, int which, int a, int b, int c,
     glfw_window = NULL;
 }
 
+- (BOOL)validateMenuItem:(NSMenuItem *)item {
+    if (item.action == @selector(performMiniaturize:)) return YES;
+    return [super validateMenuItem:item];
+}
+
+- (void)performMiniaturize:(id)sender
+{
+    if (glfw_window && (!glfw_window->decorated || glfw_window->ns.titlebar_hidden)) [self miniaturize:self];
+    else [super performMiniaturize:sender];
+}
 
 - (BOOL)canBecomeKeyWindow
 {
@@ -1595,10 +1605,10 @@ void _glfwPlatformSetWindowSizeLimits(_GLFWwindow* window,
 
 void _glfwPlatformSetWindowAspectRatio(_GLFWwindow* window, int numer, int denom)
 {
-    if (numer != GLFW_DONT_CARE && denom != GLFW_DONT_CARE)
-        [window->ns.object setContentAspectRatio:NSMakeSize(numer, denom)];
-    else
+    if (numer == GLFW_DONT_CARE || denom == GLFW_DONT_CARE)
         [window->ns.object setResizeIncrements:NSMakeSize(1.0, 1.0)];
+    else
+        [window->ns.object setContentAspectRatio:NSMakeSize(numer, denom)];
 }
 
 void _glfwPlatformSetWindowSizeIncrements(_GLFWwindow* window, int widthincr, int heightincr)
@@ -1994,8 +2004,10 @@ int _glfwPlatformCreateCursor(_GLFWcursor* cursor,
                         bitmapFormat:NSBitmapFormatAlphaNonpremultiplied
                         bytesPerRow:src->width * 4
                         bitsPerPixel:32];
-        if (rep == nil)
+        if (rep == nil) {
+            [native release];
             return false;
+        }
 
         memcpy([rep bitmapData], src->pixels, src->width * src->height * 4);
         [native addRepresentation:rep];
@@ -2060,18 +2072,38 @@ bool _glfwPlatformToggleFullscreen(_GLFWwindow* w, unsigned int flags) {
     bool made_fullscreen = true;
     bool traditional = !(flags & 1);
     NSWindowStyleMask sm = [window styleMask];
-    bool in_fullscreen = sm & NSWindowStyleMaskFullScreen;
     if (traditional) {
-        if (!(in_fullscreen)) {
-            sm |= NSWindowStyleMaskBorderless | NSWindowStyleMaskFullScreen;
-            [[NSApplication sharedApplication] setPresentationOptions: NSApplicationPresentationAutoHideMenuBar | NSApplicationPresentationAutoHideDock];
+        if (@available(macOS 10.16, *)) {
+            // As of Big Turd NSWindowStyleMaskFullScreen is no longer useable
+            if (!w->ns.in_traditional_fullscreen) {
+                w->ns.pre_full_screen_style_mask = sm;
+                [window setStyleMask: NSWindowStyleMaskBorderless];
+                [[NSApplication sharedApplication] setPresentationOptions: NSApplicationPresentationAutoHideMenuBar | NSApplicationPresentationAutoHideDock];
+                [window setFrame:[window.screen frame] display:YES];
+                w->ns.in_traditional_fullscreen = true;
+            } else {
+                made_fullscreen = false;
+                [window setStyleMask: w->ns.pre_full_screen_style_mask];
+                [[NSApplication sharedApplication] setPresentationOptions: NSApplicationPresentationDefault];
+                w->ns.in_traditional_fullscreen = false;
+            }
+            // for some reason despite calling this selector, the window doesnt actually get keyboard focus till you click on it.
+            // presumably a bug with NSWindowStyleMaskBorderless windows
+            [window performSelector:@selector(makeKeyAndOrderFront:) withObject:nil afterDelay:0];
         } else {
-            made_fullscreen = false;
-            sm &= ~(NSWindowStyleMaskBorderless | NSWindowStyleMaskFullScreen);
-            [[NSApplication sharedApplication] setPresentationOptions: NSApplicationPresentationDefault];
+            bool in_fullscreen = sm & NSWindowStyleMaskFullScreen;
+            if (!(in_fullscreen)) {
+                sm |= NSWindowStyleMaskBorderless | NSWindowStyleMaskFullScreen;
+                [[NSApplication sharedApplication] setPresentationOptions: NSApplicationPresentationAutoHideMenuBar | NSApplicationPresentationAutoHideDock];
+            } else {
+                made_fullscreen = false;
+                sm &= ~(NSWindowStyleMaskBorderless | NSWindowStyleMaskFullScreen);
+                [[NSApplication sharedApplication] setPresentationOptions: NSApplicationPresentationDefault];
+            }
+            [window setStyleMask: sm];
         }
-        [window setStyleMask: sm];
     } else {
+        bool in_fullscreen = sm & NSWindowStyleMaskFullScreen;
         if (in_fullscreen) made_fullscreen = false;
         [window toggleFullScreen: nil];
     }
@@ -2271,6 +2303,29 @@ GLFWAPI id glfwGetCocoaWindow(GLFWwindow* handle)
     return window->ns.object;
 }
 
+GLFWAPI void glfwHideCocoaTitlebar(GLFWwindow* handle, bool yes) {
+    @autoreleasepool {
+    _GLFWwindow* w = (_GLFWwindow*) handle;
+    NSWindow *window = w->ns.object;
+    w->ns.titlebar_hidden = yes;
+    NSButton *button;
+    button = [window standardWindowButton: NSWindowCloseButton];
+    if (button) button.hidden = yes;
+    button = [window standardWindowButton: NSWindowMiniaturizeButton];
+    if (button) button.hidden = yes;
+    button = [window standardWindowButton: NSWindowZoomButton];
+    [window setTitlebarAppearsTransparent:yes];
+    if (button) button.hidden = yes;
+    if (yes) {
+        [window setTitleVisibility:NSWindowTitleHidden];
+        [window setStyleMask: [window styleMask] | NSWindowStyleMaskFullSizeContentView];
+    } else {
+        [window setTitleVisibility:NSWindowTitleVisible];
+        [window setStyleMask: [window styleMask] & ~NSWindowStyleMaskFullSizeContentView];
+    }
+    } // autoreleasepool
+}
+
 GLFWAPI GLFWcocoatextinputfilterfun glfwSetCocoaTextInputFilter(GLFWwindow *handle, GLFWcocoatextinputfilterfun callback) {
     _GLFWwindow* window = (_GLFWwindow*) handle;
     _GLFW_REQUIRE_INIT_OR_RETURN(nil);
@@ -2446,7 +2501,7 @@ START_ALLOW_CASE_RANGE
 END_ALLOW_CASE_RANGE
     }
     if (utf_16_key != 0) {
-         strncpy(cocoa_key, [[NSString stringWithCharacters:&utf_16_key length:1] UTF8String], key_sz - 1);
+        strncpy(cocoa_key, [[NSString stringWithCharacters:&utf_16_key length:1] UTF8String], key_sz - 1);
     } else {
         unsigned str_pos = 0;
         for (unsigned i = 0; i < 4 && str_pos < key_sz - 1; i++) {
