@@ -161,12 +161,13 @@ coretext_all_fonts(PyObject UNUSED *_self) {
     CFArrayRef matches = CTFontCollectionCreateMatchingFontDescriptors(all_fonts_collection());
     const CFIndex count = CFArrayGetCount(matches);
     PyObject *ans = PyTuple_New(count), *temp;
-    if (ans == NULL) return PyErr_NoMemory();
+    if (ans == NULL) { CFRelease(matches); return PyErr_NoMemory(); }
     for (CFIndex i = 0; i < count; i++) {
         temp = font_descriptor_to_python((CTFontDescriptorRef) CFArrayGetValueAtIndex(matches, i));
-        if (temp == NULL) { Py_DECREF(ans); return NULL; }
+        if (temp == NULL) { CFRelease(matches); Py_DECREF(ans); return NULL; }
         PyTuple_SET_ITEM(ans, i, temp); temp = NULL;
     }
+    CFRelease(matches);
     return ans;
 }
 
@@ -212,6 +213,7 @@ manually_search_fallback_fonts(CTFontRef current_font, CPUCell *cell) {
             CFRelease(new_font);
         }
     }
+    CFRelease(fonts);
     return ans;
 }
 
