@@ -67,7 +67,12 @@
 #define MOUSE_SGR_MODE (1006 << 5)
 #define MOUSE_URXVT_MODE (1015 << 5)
 
+// Save cursor (DECSC)
+#define SAVE_CURSOR (1048 << 5)
+
 // Alternate screen buffer
+#define TOGGLE_ALT_SCREEN_1 (47 << 5)
+#define TOGGLE_ALT_SCREEN_2 (1047 << 5)
 #define ALTERNATE_SCREEN  (1049 << 5)
 
 // Bracketed paste mode
