@@ -63,12 +63,16 @@ typedef struct {
     bool close_on_child_death;
     bool window_alert_on_bell;
     bool debug_keyboard;
+    bool allow_hyperlinks;
     monotonic_t resize_debounce_time;
     MouseShape pointer_shape_when_grabbed;
+    MouseShape default_pointer_shape;
+    MouseShape pointer_shape_when_dragging;
     struct {
         UrlPrefix *values;
         size_t num, max_prefix_len;
     } url_prefixes;
+    bool detect_urls;
 } Options;
 
 typedef struct {
