@@ -50,11 +50,11 @@ class Rendering(BaseTest):
 
     def test_box_drawing(self):
         prerendered = len(self.sprites)
-        s = self.create_screen(cols=len(box_chars), lines=1, scrollback=0)
+        s = self.create_screen(cols=len(box_chars) + 1, lines=1, scrollback=0)
         s.draw(''.join(box_chars))
         line = s.line(0)
         test_render_line(line)
-        self.assertEqual(len(self.sprites), prerendered + len(box_chars))
+        self.assertEqual(len(self.sprites) - prerendered, len(box_chars))
 
     def test_font_rendering(self):
         render_string('ab\u0347\u0305你好|\U0001F601|\U0001F64f|\U0001F63a|')
@@ -128,3 +128,20 @@ class Rendering(BaseTest):
     def test_coalesce_symbol_maps(self):
         q = {(2, 3): 'a', (4, 6): 'b', (5, 5): 'b', (7, 7): 'b', (9, 9): 'b', (1, 1): 'a'}
         self.ae(coalesce_symbol_maps(q), {(1, 3): 'a', (4, 7): 'b', (9, 9): 'b'})
+        q = {(1, 4): 'a', (2, 3): 'b'}
+        self.ae(coalesce_symbol_maps(q), {(1, 1): 'a', (2, 3): 'b', (4, 4): 'a'})
+        q = {(2, 3): 'b', (1, 4): 'a'}
+        self.ae(coalesce_symbol_maps(q), {(1, 4): 'a'})
+        q = {(1, 4): 'a', (2, 5): 'b'}
+        self.ae(coalesce_symbol_maps(q), {(1, 1): 'a', (2, 5): 'b'})
+        q = {(2, 5): 'b', (1, 4): 'a'}
+        self.ae(coalesce_symbol_maps(q), {(1, 4): 'a', (5, 5): 'b'})
+        q = {(1, 4): 'a', (2, 5): 'a'}
+        self.ae(coalesce_symbol_maps(q), {(1, 5): 'a'})
+        q = {(1, 4): 'a', (4, 5): 'b'}
+        self.ae(coalesce_symbol_maps(q), {(1, 3): 'a', (4, 5): 'b'})
+        q = {(4, 5): 'b', (1, 4): 'a'}
+        self.ae(coalesce_symbol_maps(q), {(1, 4): 'a', (5, 5): 'b'})
+        q = {(0, 30): 'a', (10, 10): 'b', (11, 11): 'b', (2, 2): 'c', (1, 1): 'c'}
+        self.ae(coalesce_symbol_maps(q), {
+            (0, 0): 'a', (1, 2): 'c', (3, 9): 'a', (10, 11): 'b', (12, 30): 'a'})
