@@ -42,3 +42,7 @@ class TestTUI(BaseTest):
         self.ae(le.cursor_pos, 0)
         le.backspace()
         self.assertTrue(le.pending_bell)
+
+    def test_multiprocessing_spawn(self):
+        from kitty.multiprocessing import test_spawn
+        test_spawn()
